//----------------------------------------------------------------------------------
// File:   MultiTextureTerrain.fx
// Author: Bryan Dudash
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

float4x4 g_mWorld : World;
float4x4 g_mWorldIT : WorldInverseTranspose;
float4x4 g_mWorldViewProjection : WorldViewProjection;    

bool g_bRenderUVs = true;
bool g_bRenderWithNormalMaps = true;

cbuffer cimmutable
{
    float Kd = 0.89;
    float Ks = 0.97;
    float Kb = 1.0;
    float SpecExpon = 24.0;
    float4 ambiColor = float4(0.15,0.15,0.15,0);

    float  g_lightA1 = 5.f;
    float  g_lightA2 = 2.f;
    float3 g_lightPos = float3(3,6,-1);
    float4x4 g_Identity = {    1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};
};

//--------------------------------------------------------------------------------------
// Texture samplers
//--------------------------------------------------------------------------------------
Texture2DArray g_txTextures    : TextureArray;
Texture2DArray g_txNormals    : TextureNormalArray;

SamplerState g_samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};
SamplerState g_samAniso
{
    Filter = ANISOTROPIC;
    AddressU = Wrap;
    AddressV = Wrap;
};
SamplerState g_samPoint
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Wrap;
    AddressV = Wrap;
};

//--------------------------------------------------------------------------------------
// Vertex shader output structure
//--------------------------------------------------------------------------------------

struct VS_to_PS
{
    float4 pos          : SV_Position;
    float4 tex0         : TEXTURE0;
    float4 tex1         : TEXTURE1;
    float3 norm         : TEXTURE3;
    float3 tangent      : TEXTURE4;
    float3 binorm       : TEXTURE5;
    float3 worldPos     : TEXTURE6;    
};

VS_to_PS RenderSceneVS( 
                        // regular mesh data
                        float3 vPos            : POSITION,
                        float3 vNormal        : NORMAL,
                        float3 vTangent        : TANGENT,
                        float3 vBiNormal    : BINORMAL,
                        float4 vTexCoord0    : TEXCOORD0,
                        float4 vTexCoord1    : TEXCOORD1
                        )
{
    VS_to_PS output;
    
    // Transform the position from object space to homogeneous projection space
    output.pos = mul(float4(vPos.xyz,1) , g_mWorldViewProjection);
    
    // tangent basis (world space) for lighting
    output.norm.xyz = normalize(mul(vNormal, (float3x3)g_mWorldIT)); 
    output.tangent.xyz = normalize(mul(vTangent, (float3x3)g_mWorldIT));     
    output.binorm.xyz = normalize(mul(vBiNormal, (float3x3)g_mWorldIT)); 

    // world position for lighting
    output.worldPos.xyz = mul(float4(vPos.xyz,1), g_mWorld);

    // just pass thru our tex coords
    output.tex0 = vTexCoord0;
    output.tex1 = vTexCoord1;
    
    return output;
}


float4 RenderScenePS( VS_to_PS input) : SV_Target
{
    // Load the color values (3rd component is the slice of the aray from which to sample)
    // We multple by the last component of the texcoord, which is set by the application to be the blend weight
    float4 albedo0;
    if(g_bRenderUVs)
        albedo0 = float4(1,0,0,1)*input.tex0.w;
    else
        albedo0 = g_txTextures.Sample(g_samAniso,input.tex0.xyz)*input.tex0.w;
    
    float4 albedo1;
    if(g_bRenderUVs)    
        albedo1 = float4(0,1,0,1)*input.tex1.w;
    else
        albedo1 = g_txTextures.Sample(g_samAniso,input.tex1.xyz)*input.tex1.w;
           
    float4 bumps0 = g_txNormals.Sample(g_samAniso,input.tex0.xyz)*input.tex0.w;    
    float4 bumps1 = g_txNormals.Sample(g_samAniso,input.tex1.xyz)*input.tex1.w;    

    // combine using sampled alphas
    float3 finalAlbedo = (albedo0.xyz * albedo0.w) + (albedo1.xyz * albedo1.w);
    float3 finalBumps = bumps0.xyz + bumps1.xyz;

    finalBumps = normalize(finalBumps);
    
    // lighting
    float3 Nn = normalize(input.norm);
    float3 Tn = normalize(input.tangent);
    float3 Bn = normalize(input.binorm);
    float3 Nb = Nn;
    if(g_bRenderWithNormalMaps) 
        Nb += (finalBumps.x * Tn + finalBumps.y * Bn);
    Nb = normalize(Nb);
    float3 toLight = g_lightPos - input.worldPos.xyz;
    float dist = length(toLight);
    float3 Ln = toLight / dist;
    float ldn = max(0.0,dot(Ln,Nb));
    float4 diffContrib = saturate(Kd*(ldn) + ambiColor);
    float scale = saturate(g_lightA1/dist + g_lightA2 / (dist*dist));
    diffContrib *= scale;
    float4 result = float4(finalAlbedo*diffContrib.xyz,1);    
    return result;
}

float4 ZOnlyPS( VS_to_PS input) : SV_Target
{
    return float4(0,0,0,1);
}


RasterizerState SolidRS
{
    FillMode = Solid;
	MultisampleEnable = TRUE;
};

BlendState ZOnly
{
    RenderTargetWriteMask[0] = 0;
};

BlendState Normal
{
    RenderTargetWriteMask[0] = 0x0F;
};

DepthStencilState ZwriteNoZtest
{
    DepthEnable = false;
};

DepthStencilState NormalDS
{
    DepthEnable = true;
};

//--------------------------------------------------------------------------------------
technique10 RenderTerrain
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, RenderSceneVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, RenderScenePS() ) );
        SetBlendState( Normal, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState(NormalDS,0);
        SetRasterizerState(SolidRS);
    }
}

technique10 RenderTerrainZOnly
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, RenderSceneVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, ZOnlyPS() ) );
        SetBlendState( ZOnly, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState(ZwriteNoZtest,0);
        SetRasterizerState(SolidRS);
    }
}
